﻿using System.Collections;
using System.Threading;
using Unity.Jobs;
using UnityEngine;

/// <summary>
/// Divide the terrain generation into 9 "chunks".
/// Each chunk is responsible for the generation of its
/// terrain object and only its terrain object.
/// 
/// A chunk may be moved when the player moves, such that a 3x3
/// grid of chunks surrounds the player at any time, and that
/// there's a minimum of one full terrain between the player and
/// the void.
/// 
/// Once a chunk is moved, it will regenerate its terrain object.
/// This will be done by a thread/job.
/// </summary>
public class TerrainChunker : MonoBehaviour
{
    [Header("Autogenerated values")]
	[Tooltip("Current grid position of the chunker")]
    public Vector2Int gridPosition;

    public Vector2 crawlOffset;

    public TerrainMaster master;

    private Terrain terrain;

    private TerrainData data;

    private int heightResolution = 513;

    private float terrainWidth = 1000;

    private float heightScale;

    private ShapeGeneratorJob shapeGenerator;
    private JobHandle runningShape;
    private bool lookForCompletion = false;

    public void Initialize(TerrainMaster master, Vector2Int gridPosition,
        Vector2 crawlOffset)
    {
        this.master = master;
        this.gridPosition = gridPosition;
        this.crawlOffset = crawlOffset;

        gameObject.name = $"Chunker ({gridPosition})";

        data = Instantiate(Resources.Load<TerrainData>("Template"));

        heightScale = data.size.y / master.terrainHeight;
        data.size = new(data.size.x, master.terrainHeight, data.size.z);

        var terrainGO = Terrain.CreateTerrainGameObject(data);
        terrainGO.transform.parent = transform;

        heightResolution = data.heightmapResolution;
        terrainWidth = data.size.x;

        UpdatePositions(gridPosition, true);
    }

    public void UpdatePositions(Vector2Int newGridPosition, bool forceUpdate = false)
    {
        print($"Updating {gameObject.name} position.");

        if (newGridPosition != gridPosition || forceUpdate)
        {
            gridPosition = newGridPosition;

            Vector2 worldPosition = (Vector2)gridPosition * terrainWidth;
            transform.position = new(worldPosition.x, 0, worldPosition.y);

            // Now we update the terrain
            UpdateTerrain();
        }
    }

    private void UpdateTerrain()
    {
        print($"Updating {gameObject.name} terrain.");
        if (!runningShape.IsCompleted)
        {
            // Cancel thread if its running.
            shapeGenerator.Cancel();
            print($"{gameObject.name} has running job, which has been cancelled.");
        }

        shapeGenerator = new(crawlOffset, gridPosition, heightResolution,
            heightScale, master.octaves);
        runningShape = shapeGenerator.Schedule();
        lookForCompletion = true;
    }

    private void ApplyTerrain(float[,] heights)
    {
        print($"Apply {gameObject.name} terrain.");
        data.SetHeights(0, 0, heights);
    }

    private void Update()
    {
        if (lookForCompletion && runningShape.IsCompleted)
        {
            runningShape.Complete();
            ApplyTerrain(shapeGenerator.Retrieve());
            lookForCompletion = false;
        }
    }

    private void OnDestroy()
    {
        shapeGenerator.Cancel();
    }
}